import numpy as np
import google.cloud.firestore as firestore
from txp_cloud.common.config import settings
from typing import List
import logging
from enum import Enum

configurations_collection_name = settings.firestore.configurations_collection
configuration_reference_field_path = "configuration_ref"
configuration_id_field = "configuration_id"
log = logging.getLogger(__name__)
edges_collection_name = settings.firestore.edges_collection


class SignalMode(Enum):
    TIME = 0
    TIME_FFT = 1
    TIME_PSD = 2
    TIME_FFT_PSD = 3
    FFT = 4
    FFT_PSD = 5
    PSD = 6
    IMAGE = 7

    @staticmethod
    def is_time_only(mode):
        return not SignalMode.is_fft(mode) and not SignalMode.is_psd(mode)

    @staticmethod
    def is_time(mode):
        return mode in [SignalMode.TIME.value, SignalMode.TIME_FFT.value, SignalMode.TIME_FFT_PSD.value,
                        SignalMode.TIME_PSD.value,
                        SignalMode.IMAGE.value]

    @staticmethod
    def is_fft(mode):
        return mode in [SignalMode.TIME_FFT.value, SignalMode.TIME_FFT_PSD.value, SignalMode.FFT.value,
                        SignalMode.FFT_PSD.value]

    @staticmethod
    def is_psd(mode):
        return mode in [SignalMode.TIME_PSD.value, SignalMode.TIME_FFT_PSD.value, SignalMode.PSD.value,
                        SignalMode.FFT_PSD.value]


def get_sampling_frequency(signal, client):
    configuration = pull_configuration_from_firestore(client, signal["configuration_id"])
    documents = client.collection(edges_collection_name). \
        where(configuration_reference_field_path, "==", configuration.reference). \
        where("logical_id", "==", signal["edge_logical_id"]).get()
    if len(documents) != 1:
        return 1
    edge_document = documents[0]
    return edge_document.get("is_of_kind_ref").get().get("perceptions")[signal["perception_name"]]["sampling_frequency"]


def from_proto_to_json(proto):
    timestamp, index = proto.metadata.package_id.split("_")
    timestamp = int(timestamp)
    index = int(index)
    prev_index = proto.metadata.previous_part_index
    res = []
    for signal in proto.signals:
        row = {
            "data": [],
            "package_timestamp": timestamp,
            "signal_timestamp": signal.timestamp,
            "previous_part_index": prev_index,
            "part_index": index,
            "perception_name": signal.perception_name,
            "edge_logical_id": proto.metadata.edge_descriptor.logical_id,
            "configuration_id": proto.configuration_id,
            "label": "OK"
        }
        for i, dimension_signal_sample in enumerate(signal.samples):
            row["data"].append({
                "values": list(dimension_signal_sample.samples),
                "index": i
            })
            i += 1
        res.append(row)
    return res


def merge_signal_chunks(signal_chunks):
    """Merges all signal chunks for a given signal row.

    Args:
        signal_chunks: chunks of a signal, the signal is split in chunks due to MQTT maximum size per package

    Return:
        merged element.
    """

    data = []
    for i in range(len(signal_chunks[0]["data"])):
        data.append([])
    for i in range(0, len(signal_chunks)):
        for j, dimension_signal_sample in enumerate(signal_chunks[i]["data"]):
            data[j] += dimension_signal_sample["values"]
    signal_chunks[0]["data"] = [{"values": dimension, "index": i} for i, dimension in enumerate(data)]
    signal_chunks[0]["previous_part_index"] = 0
    return signal_chunks[0]


def get_fft_as_np_array(bigquery_fft):
    fft_per_dimensions = []
    for dimension in bigquery_fft:
        fft_dimension = []
        for z in dimension["values"]:
            fft_dimension.append(np.complex128(complex(z["real"], z["imag"])))
        fft_per_dimensions.append(np.array(fft_dimension))

    return np.array(fft_per_dimensions)


def pull_current_configuration_from_firestore(db: firestore.Client) -> firestore.DocumentSnapshot:
    """Pull down from the Firestore DB the most recent configuration document.

    Returns:
        The configuration firestore.DocumentReference.
    """
    configuration = db.collection(configurations_collection_name).order_by(
        "server_timestamp", direction=firestore.Query.DESCENDING
    ).limit(1).get()

    if not configuration:
        log.warning("No Configuration document was found.")
        return None

    return configuration[0]


def pull_configuration_from_firestore(db: firestore.Client, configuration_id) -> firestore.DocumentSnapshot:
    """Pull down from the Firestore DB the most recent configuration document.

    Returns:
        The configuration firestore.DocumentReference.
    """
    configuration = db.collection(configurations_collection_name).where(configuration_id_field, "==",
                                                                        configuration_id).limit(1).get()

    if not configuration:
        log.warning("No Configuration document was found.")

    return configuration[0]


def pull_docs_from_collection_associated_with_configuration(db: firestore.Client,
                                                            collection_name: str,
                                                            configuration_ref: firestore.DocumentReference) -> List[
    firestore.DocumentSnapshot]:
    """Returns the documents of a collection that has the configuration value equal to
    the configuration reference received.

    Returns:
        The list of documents found for the query.
    """
    documents = db.collection(collection_name).where(configuration_reference_field_path, "==", configuration_ref).get()

    if not documents:
        log.warning(f"No documents found in collection {collection_name} for configuration {configuration_ref.id}")

    return documents


def get_signal_mode_from_firestore(config_id, edge_logical_id, perception_name, client):
    configuration = pull_configuration_from_firestore(client, config_id)
    documents = client.collection(edges_collection_name). \
        where(configuration_reference_field_path, "==", configuration.reference). \
        where("logical_id", "==", edge_logical_id).get()
    if len(documents) != 1:
        return None
    edge_document = documents[0]
    return edge_document.get("is_of_kind_ref").get().get("perceptions")[perception_name]["mode"]



